/* Copyright (C)  Walter L. Peacock 1984-1988.  All rights reserved */

/* THIS PROGRAM BELONGS TO WALTER L. PEACOCK.  IT IS CONSIDERED A TRADE */
/* SECRET AND IS NOT TO BE DIVULGED OR USED BY PARTIES WHO HAVE NOT */
/* RECEIVED WRITTEN AUTHORIZATION FROM THE OWNER. */

/* 06/17/87   btsetup.c  */
/*  Maintain btparms.btr file containing parameter data on the b-trees. */
/*  btsetup.c will maintain up to MAXBTREE entries in the btparms.btr file. */
/*  If you need more than 20 entries in btparms.btr then increase MAXBTREE. */
/*  MAXBTREE is defined in cbtree.h and is only used by btrinit.c and	*/
/*  btsetup.c */
/*  MAXFNAME is defined in cbtree.h and defines the maximum length of the */
/*  index file name, including PATH.  This can be increased as needed. */
/*  Increasing MAXFNAME will increase the amount of memory required by each */
/*  btcommo structure.  The default value is 12. */

#include <stdio.h>
#include "cbtree.h"

#if CI | DC
#else
#include <ctype.h>
#endif

#if MSDOS & LC
int _stack = 4000;
#endif

#if AmigaDOS & LC
#define printf iprintf   /* use fast integer printf */
#endif

#define MAXDEPTH   3   /* assume a maximum cbtree depth of 3 for DB size   */
#define MAXNAME   8    /* maximum # characters in B+tree name */
#define MAXNFLD   5    /* maximum # characters in numerical fields */
#define SCRNLEN   20   /* maximum # lines that can be displayed on the screen */

/* ENTRY STRUCTURE for storing line buffer data */
#define BT_ENTRY struct _bt_entry
BT_ENTRY
{
   char parmname[MAXNAME+1];   /* Parameter name      */
   char indxname[MAXFNAME+1];   /* Index file name */
   char btrnumb[MAXNFLD+1];    /* B+tree Number */
   char ilen[MAXNFLD+1];       /* Index Block Size   */
   char cell[MAXNFLD+1];       /* Cells per Index Blk  */
   char klen[MAXNFLD+1];       /* Key Length         */
   char dlen[MAXNFLD+1];       /* Data Record Length   */
   char szmax[MAXNFLD+1];      /* sizeof btrecnum      */
   char dbszbuf[12];           /* data base size */
};

main()
{
   extern FILE *fopen();
   extern long lpow();
   extern char *strnncpy(), *strcpy(), *strcat(), *calloc(), *fgets();
   extern char *delicpy();   /* copy up to delimiter */
   extern void  free(), exit(), pr_hdr(), pr_en(), load_en();
   extern uint  edit_en(); /* edit globals and store in line buffer */
   extern int   atoi();
   BT_ENTRY  en[MAXBTREE];   /* array of entry structures */
   BT_ENTRY *enp;
   long  dbsz;
   FILE *btfp;
   int   contflag = 0;
   int   btploc, btsel, scrncnt;
   uint  cellcnt;
   char  line[MAXBTREE][LINWIDTH + 1];
   char  sel[20];
   char  c;
   register int i;

   wopen("CON:0/0/640/200/Btsetup");

   /***   first read in the entire file and load array of lines[]s ***/
   if( (btfp = fopen("btparms.btr","r") ) == NULL)   /* read only */
      if( (btfp = fopen("btparms.btr","a+") ) == NULL)   /* read only */
         ckerror(- CK_OPEN, "btsetup: btparms.btr");

   memset(line, NUL, sizeof(line));
   memset(en, NUL, sizeof(en));

   for (i = 0; i < MAXBTREE; ++i)   /* read in all lines   */
   {
      char *cp;

      cp = fgets(line[i], LINWIDTH, btfp);
      if (cp == NULL || *cp == '\n')   /* read the line */
         break;
      load_en(en + i, line[i]);
   }

   fclose(btfp);

   while (!contflag)
   {
	scrncnt = SCRNLEN;

      repeat
      {
         pr_hdr();   /* clear screen and print the table header */

         /* display all the fields on the screen */
         for (btploc = 0; btploc < MAXBTREE; ++btploc) /* scan for all entries */
         {

		 	if(btploc >= scrncnt){
				scrncnt += SCRNLEN;
                scr_curs(SCRNLEN + 2, 0);
				printf("Press ENTER to continue.");
				gets(sel);
                pr_hdr();   /* clear screen and print the table header */
				/***********************
                scr_curs(SCRNLEN + 2, 0);
				printf("                          ");
				***********************/
			}

            enp = en + btploc;   /* entry pointer */
            if (*enp->parmname)  /* line has stuff on it */
            {

               /* reckon the data base size */
               cellcnt = (uint)atoi(enp->cell);  /* reckon the cell count */
               dbsz = lpow((cellcnt+1)>>1, MAXDEPTH);   /* LC bug */
               if (*enp->parmname != NUL)
                  sprintf(enp->dbszbuf, "%ld",  dbsz);
               else
                  *enp->dbszbuf = NUL;
            }
            else
               *enp->parmname = *enp->indxname = *enp->btrnumb = *enp->ilen =
                     *enp->klen = *enp->dlen = *enp->szmax = *enp->dbszbuf =
                     *enp->cell = NUL;

            /* print out the entry */
            scr_curs(btploc + SCRNLEN - scrncnt + 1, 0);
            pr_en(btploc, enp);
            for (i = btploc + 1; i < MAXBTREE; ++i)
            {
               if (*en[i].parmname &&
                     strcmp(enp->indxname, en[i].indxname) == 0 &&
                     strcmp(enp->ilen, en[i].ilen) != 0)
               {
                  scr_curs(SCRNLEN + 2, 0);
                  printf("\7'%s' and '%s' have DIFFERENT block sizes, PLEASE FIX!",
                        enp->parmname, en[i].parmname);
                  break;
               }
            }
         }

         printf("\n\nEnter B+tree ## (or [S]ave, [Q]uit or [D]elete) ==> _");
         repeat
         {
            scr_curs(SCRNLEN + 3, 52);
            gets(sel);

            if ((c = toupper(*sel)) == 'Q'  ||  c == 'S'  ||  c == 'D')
            {
               if (c == 'D')
               {
                  scr_curs(SCRNLEN + 2, 0);
                  printf("Which entry would you like to DELETE ? ==> __\b\b");
                  gets(sel);
                  if ( (btsel = atoi(sel) - 1) < 0  ||  btsel >= MAXBTREE)
                  {
                     btsel = -1;
                     break;
                  }
                  blkshft(line[btsel], line[btsel+1], MAXBTREE-1-btsel,
                        sizeof(line[0]));
                  blkshft(en + btsel, en + btsel+1, MAXBTREE-1-btsel,
                        sizeof(en[0]));
                  strcpy(line[MAXBTREE-1], "\n");   /* clear line */
                  *en[MAXBTREE-1].parmname = NUL;    /* clear entry */
                  btsel = -1;           /* force a redisplay */
                  break;
               }
               else if (c == 'S')
               {
                  /***   now write out the file   ***/
                  if( (btfp = fopen("btparms.btr","w") ) == NULL)   /* read only */
                     ckerror(- CK_OPEN, "btsetup: btparms.btr");

                  for (i = 0; i < MAXBTREE; ++i)   /* all lines  */
                     fputs(line[i], btfp);  /* store the line in the array   */

                  fclose(btfp);
               }
               goto set_0exit;
            }
            else
               btsel = atoi(sel) - 1;
         }
         until( (btsel >= 0)   &&   (btsel < MAXBTREE) );
      }
      until(btsel != -1);   /* make sure a DELETE error hasn't occurred */

      /**** edit the selected record ****/
      contflag = edit_en(btsel, en + btsel, line[btsel]);   /* edit & re-store in line */

   }      /* end while(!contflag) */

set_0exit:
   wclose();
   return(0);   /* make MW CSD happy */
}

/*******************/

void load_en(en, linebuf)   /* load ENTRY structure from line */

BT_ENTRY *en;
char     *linebuf;

{
   extern char *delicpy(), *strcpy(), *strcat();   /* copy to delimiter */
   char *fbuff;

   fbuff = delicpy(en->parmname, linebuf, MAXNAME);   /* B+tree name      */
   fbuff = delicpy(en->ilen,   fbuff, MAXNFLD);       /* Index Block Sz   */
   fbuff = delicpy(en->klen,   fbuff, MAXNFLD);       /* Key Length       */
   fbuff = delicpy(en->cell,   fbuff, MAXNFLD);       /* Cells / Idx Blk  */
   fbuff = delicpy(en->dlen,   fbuff, MAXNFLD);       /* Data Rec Len     */
   fbuff = delicpy(en->szmax,  fbuff, MAXNFLD);       /* sizeof btrecnum  */
   fbuff = delicpy(en->indxname,fbuff, MAXFNAME);      /* index file name  */
   if (*en->indxname == '\n')
   {
      strcpy(en->indxname, en->parmname);
      strcat(en->indxname, ".idx");
      strcpy(en->btrnumb, "0");
   }
   else
   {
      fbuff = delicpy(en->btrnumb, fbuff, MAXNFLD);   /* B+tree Number */
   }
}

/***************************/

uint edit_en(btploc, enp, linebuf)   /* edit globals and store in line buffer */

int       btploc;
BT_ENTRY *enp;
char     *linebuf;

{
   extern char *strnncpy(), *strcpy(), *strcat();
   extern long  lpow();      /* raise x to the power y */
   extern int   atoi();
   extern void  pr_hdr(), pr_en();
   long      dbsz;
   uword cellcnt;
   uword kylen;
   uword idxsz;
   uint u;
   char sel[20];

   for(;;)                  /* until they are satisfied with the edit */
   {

      /* reckon the data base size */
      cellcnt = (uword)atoi(enp->cell);
      dbsz = lpow((cellcnt+1)>>1, MAXDEPTH);   /* Lattice needs '>>' */
      if (*enp->parmname != NUL)
         sprintf(enp->dbszbuf, "%ld",  dbsz);
      else
         *enp->dbszbuf = NUL;

      pr_hdr();   /* clear screen and print the table header */
      pr_en(btploc, enp);
      printf("\n\nEnter <CR> to edit this entry, or 'R' to return ==> _\b");
      gets(sel);
      if(toupper(*sel) == 'Q')
         return (1);

      if (toupper(*sel) == 'R')
         return (0);

      printf("\n\nEnter B+tree parameters below or <CR> if no change.\n");

      printf("\n B+tree Name: ________\b\b\b\b\b\b\b\b");
      printf("%s", enp->parmname);
      scr_curs(9, 14);
      gets(sel);
      scr_curs(9, 14+strlen(sel));
      if (*sel != NUL)
      {
         printf("       \r");
         if (*enp->parmname == NUL)   /* initialize the rest too */
         {
            strnncpy(enp->indxname, sel, MAXNAME);
            strcat(enp->indxname, ".idx");
            strcpy(enp->btrnumb, "0");
         }
         strnncpy(enp->parmname, sel, MAXNAME);
      }

      printf("\n Index File Name: ____________\b\b\b\b\b\b\b\b\b\b\b\b");
      printf("%s", enp->indxname);
      scr_curs(10, 18);
      gets(sel);
      scr_curs(10, 18+strlen(sel));
      if (*sel != NUL)
      {
         printf("           \r");
         strnncpy(enp->indxname, sel, MAXFNAME);
      }

      printf("\n (Leave zero) : ___\b\b\b");
      printf("%s", enp->btrnumb);
      scr_curs(11, 16);
      gets(sel);
      scr_curs(11, 16+strlen(sel));
      if (*sel != NUL)
      {
         printf("  \r");
         strnncpy(enp->btrnumb, sel, MAXNFLD);
      }

      printf("\n Index Block Size: _____\b\b\b\b\b");
      printf("%s", enp->ilen);
      scr_curs(12, 19);
      gets(sel);
      scr_curs(12, 19+strlen(sel));
      if (*sel != NUL)
      {
         printf("    \r");
         strnncpy(enp->ilen, sel, MAXNFLD);
      }

      printf("\n Key Length: ____\b\b\b\b");
      printf("%s", enp->klen);
      scr_curs(13, 13);
      gets(sel);
      scr_curs(13, 13+strlen(sel));
      if (*sel != NUL)
      {
         printf("   \r");
         strnncpy(enp->klen, sel, MAXNFLD);
      }

      printf("\n Data Record Size: _____\b\b\b\b\b");
      printf("%s", enp->dlen);
      scr_curs(14, 19);
      gets(sel);
      scr_curs(14, 19+strlen(sel));
      if (*sel != NUL)
      {
         printf("    \r");
         strnncpy(enp->dlen, sel, MAXNFLD);
      }

      printf("\n Maximum Records in btrecnum: ____\b\b\b\b");
      printf("%s", enp->szmax);
      scr_curs(15, 30);
      gets(sel);
      scr_curs(15, 30+strlen(sel));
      if (*sel != NUL)
      {
         printf("\   \r");
         strnncpy(enp->szmax, sel, 4);
      }

      /* reckon the cell count */
      kylen = (uword)atoi(enp->klen);
      idxsz = (uword)atoi(enp->ilen);
      if (kylen > 0   &&  idxsz > 0)
      {
         u = kylen + 3 * sizeof(long) + 2 * sizeof(uword) + 3 * SZCHAR;
         if (idxsz < u || idxsz < 64) /* absolute minimum size index block */
         {
            printf("\n\n\tIndex Block Size too small: ");
            printf("\n\tYou may:");
            printf("\n\t\to Increase the Index Block size, or");
            printf("\n\t\to Reduce the Key Length.");
            gets(sel);
         }
         cellcnt = (idxsz - 2 * sizeof(long) - 2 * sizeof(uword) - 2) /
               (sizeof(long) + kylen + SZCHAR);
         sprintf(enp->cell, "%u",   cellcnt);
      }

      strcpy(linebuf, enp->parmname);
      strcat(linebuf, "^");
      strcat(linebuf, enp->ilen);
      strcat(linebuf, "^");
      strcat(linebuf, enp->klen);
      strcat(linebuf, "^");
      strcat(linebuf, enp->cell);
      strcat(linebuf, "^");
      strcat(linebuf, enp->dlen);
      strcat(linebuf, "^");
      strcat(linebuf, enp->szmax);
      strcat(linebuf, "^");
      strcat(linebuf, enp->indxname);
      strcat(linebuf, "^");
      strcat(linebuf, enp->btrnumb);
      strcat(linebuf, "^");
      strcat(linebuf, "\n");
   }   /* end for(;;) */

#if LC
   return (1);   /* prevent Lattice warning */
#endif
}

/*************/

long lpow(x, y)   /* x to the power of y */
int x;
uint y;
{
   long l;

   l = x;                 /* upgrade to a long */
   while (--y != 0)
      l *= x;
   return (l);
}

/***********/

void pr_hdr()   /* print out the header */

{
   scr_clr();                                          /* clear screen   */
   printf("Btr  B+tree   Index File         Block  Key   Data    Max  Cells/  DB \n");
   printf("Num  Name       Name       TBA   Size   Len  Rec Sz  Recs  Block  size");
}

/*****************/

void pr_en(btploc, en)   /* print globals as a line */

int       btploc;
BT_ENTRY *en;      /* line number */

{
   int i;
   
	/* display up to last 12 chars of index file name. */
	if( (i = strlen(en->indxname)) <= 12)
		i = 0;
	else
		i -= 12;

   printf("\n%3d", btploc + 1);
   if (*en->parmname)
      printf(" %-8.8s %-12.12s  %2.2s   %5.5s  %4.4s  %5.5s   %4.4s  %4.4s   %s",
            en->parmname, &en->indxname[i], en->btrnumb, en->ilen, en->klen,
            en->dlen, en->szmax, en->cell, en->dbszbuf);
}

/************************/

char *delicpy(dst, src, n)   /* copy up to n chars or to delimiter */
                             /* return pointer to char AFTER delimiter */
char *dst, *src;
uint n;
{
   char c;

   while (n-- != 0)            /* do up to n characters */
   {
      if ((c = *dst++ = *src++) == DELIMITER  ||  c == '\n')
      {
         *--dst = NUL;   /* replace delimiter with NUL */
         return(src);   /* return pointer to char AFTER delimiter */
      }
      else if (c == NUL)
         return (--src);   /* point to the NUL */
   }
   *dst = NUL;               /* place a NUL after n bytes */
   while (*src++ != DELIMITER)   /* eat up leftovers */
      ;
   return(src);   /* return pointer to char AFTER delimiter */
}

/* Copyright (C)  Walter L. Peacock 1984-1988.  All rights reserved */

/* THIS PROGRAM BELONGS TO WALTER L. PEACOCK.  IT IS CONSIDERED A TRADE */
/* SECRET AND IS NOT TO BE DIVULGED OR USED BY PARTIES WHO HAVE NOT      */
/* RECEIVED WRITTEN AUTHORIZATION FROM THE OWNER.                        */
